# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = 'ml_collections'
copyright = '2020, The ML Collection Authors'
author = 'The ML Collection Authors'

# The full version, including alpha/beta/rc tags
release = '0.1.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'nbsphinx',
    'recommonmark',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

autosummary_generate = True

master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Config file where `get_config` takes a string argument."""

import ml_collections


def get_config(config_string):
  """A config which takes an extra string argument."""
  possible_configs = {
      'type_a': ml_collections.ConfigDict({
          'thing_a': 23,
          'thing_b': 42,
      }),
      'type_b': ml_collections.ConfigDict({
          'thing_a': 19,
          'thing_c': 65,
      }),
  }
  return possible_configs[config_string]



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Example of ConfigDict usage.

This example includes loading a ConfigDict in FLAGS, locking it, type
safety, iteration over fields, checking for a particular field, unpacking with
`**`, and loading dictionary from string representation.
"""

from absl import app
from ml_collections.config_flags import config_flags
import yaml

_CONFIG = config_flags.DEFINE_config_file(
    'my_config',
    default='ml_collections/config_dict/examples/config.py')


def dummy_function(string, **unused_kwargs):
  return 'Hello {}'.format(string)


def print_section(name):
  print()
  print()
  print('-' * len(name))
  print(name.upper())
  print('-' * len(name))
  print()


def main(_):
  # Config is already loaded in FLAGS.my_config due to the logic hidden
  # in app.run().
  config = _CONFIG.value

  print_section('Printing config.')
  print(config)

  # Config is of our type ConfigDict.
  print('Type of the config {}'.format(type(config)))

  # By default it is locked, thus you cannot add new fields.
  # This prevents you from misspelling your attribute name.
  print_section('Locking.')
  print('config.is_locked={}'.format(config.is_locked))
  try:
    config.object.new_field = -3
  except AttributeError as e:
    print(e)

  # There is also "did you mean" feature!
  try:
    config.object.floet = -3.
  except AttributeError as e:
    print(e)

  # However if you want to modify it you can always unlock.
  print_section('Unlocking.')
  with config.unlocked():
    config.object.new_field = -3
    print('config.object.new_field={}'.format(config.object.new_field))

  # By default config is also type-safe, so you cannot change the type of any
  # field.
  print_section('Type safety.')
  try:
    config.float = 'jerry'
  except TypeError as e:
    print(e)
  config.float = -1.2
  print('config.float={}'.format(config.float))

  # NoneType is ignored by type safety and can both override and be overridden.
  config.float = None
  config.float = -1.2

  # You can temporarly turn type safety off.
  with config.ignore_type():
    config.float = 'tom'
    print('config.float={}'.format(config.float))
    config.float = 2.3
    print('config.float={}'.format(config.float))

  # You can use ConfigDict as a regular dict in many typical use-cases:
  # Iteration over fields:
  print_section('Iteration over fields.')
  for field in config:
    print('config has field "{}"'.format(field))

  # Checking if it contains a particular field using the "in" command.
  print_section('Checking for a particular field.')
  for field in ('float', 'non_existing'):
    if field in config:
      print('"{}" is in config'.format(field))
    else:
      print('"{}" is not in config'.format(field))

  # Using ** unrolling to pass the config to a function as named arguments.
  print_section('Unpacking with **')
  print(dummy_function(**config))

  # You can even load a dictionary (notice it is not ConfigDict anymore) from
  # a yaml string representation of ConfigDict.
  # Note: __repr__ (not __str__) is the recommended representation, as it
  # preserves FieldReferences and placeholders.
  print_section('Loading dictionary from string representation.')
  dictionary = yaml.load(repr(config), yaml.UnsafeLoader)
  print('dict["object_reference"]["dict"]["dict"]["float"]={}'.format(
      dictionary['object_reference']['dict']['dict']['float']))


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Custom parser to override tuples in the config dict."""
import ast

from absl import flags


class TupleParser(flags.ArgumentParser):
  """Parser for tuple arguments.

  Custom flag parser for Tuple objects that is based on the existing parsers
  in `absl.flags`. This parser can be used to read in and override tuple
  arguments. It outputs a `tuple` object from an existing `tuple` or `str`.
  The ony requirement is that the overriding parameter should be a `tuple`
  as well. The overriding parameter can have a different number of elements
  of different type than the original. For a detailed list of what `str`
  arguments are supported for overriding, look at `ast.literal_eval` from the
  Python Standard Library.
  """

  def parse(self, argument):
    """Returns a `tuple` representing the input `argument`.

    Args:
      argument: The argument to be parsed. Valid types are `tuple` and
        `str`. An empty `tuple` is returned for arguments `NoneType`.

    Returns:
      A `TupleType` representing the input argument as a `tuple`.

    Raises:
      `TypeError`: If the argument is not of type `tuple`, `str`, or
        `NoneType`.
      `ValueError`: If the string is not a well formed `tuple`.
    """
    if isinstance(argument, tuple):
      return argument
    elif isinstance(argument, str):
      return _convert_str_to_tuple(argument)
    elif argument is None:
      return ()
    else:
      msg = ('Could not parse argument {} of type '
             '{} for element of type `tuple`.'
            ).format(argument, type(argument))
      raise TypeError(msg)

  def flag_type(self):
    return 'tuple'


def _convert_str_to_tuple(string):
  """Function to convert a Python `str` object to a `tuple`.

  Args:
    string: The `str` to be converted.

  Returns:
    A `tuple` version of the string.

  Raises:
    ValueError: If the string is not a well formed `tuple`.
  """
  # literal_eval converts strings to int, tuple, list, float and dict,
  # booleans and None. It can also handle nested tuples.
  # It does not, however, handle elements of type set.
  try:
    value = ast.literal_eval(string)
  except ValueError:
    # A ValueError is raised by literal_eval if the string is not well
    # formed. Catch it and print out a more readable statement.
    msg = 'Argument {} does not evaluate to a `tuple` object.'.format(string)
    raise ValueError(msg)
  except SyntaxError:
    # The only other error that may be raised is a `SyntaxError` because
    # `literal_eval` calls the Python in-built `compile`. This error is
    # caused by parsing issues.
    msg = 'Error while parsing string: {}'.format(string)
    raise ValueError(msg)

  # Make sure we got a tuple. If not, its an error.
  if isinstance(value, tuple):
    return value
  else:
    raise ValueError('Expected a tuple argument, got {}'.format(type(value)))



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Example of basic ConfigDict usage.

This example shows the most basic usage of ConfigDict, including type safety.
For examples of more features, see example_advanced.
"""

from absl import app
import ml_collections


def main(_):
  cfg = ml_collections.ConfigDict()
  cfg.float_field = 12.6
  cfg.integer_field = 123
  cfg.another_integer_field = 234
  cfg.nested = ml_collections.ConfigDict()
  cfg.nested.string_field = 'tom'

  print(cfg.integer_field)  # Prints 123.
  print(cfg['integer_field'])  # Prints 123 as well.

  try:
    cfg.integer_field = 'tom'  # Raises TypeError as this field is an integer.
  except TypeError as e:
    print(e)

  cfg.float_field = 12  # Works: `int` types can be assigned to `float`.
  cfg.nested.string_field = u'bob'  # `String` fields can store Unicode strings.

  print(cfg)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Config file that raises TypeError on import.

When trying loading the configuration file as a flag, the flags library catches
TypeError exceptions then recasts them as a IllegalFlagTypeError and rethrows
(b/63877430). The rethrow does not include the stacktrace from the original
exception, so we manually add the stracktrace in configflags.parse(). This is
tested in `ConfigFlagTest.testTypeError` in `config_overriding_test.py`.
"""


def type_error_function():
  raise TypeError('This is a TypeError.')


def get_config():
  return {'item': type_error_function()}



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Config flags module."""

from .config_flags import DEFINE_config_dataclass
from .config_flags import DEFINE_config_dict
from .config_flags import DEFINE_config_file
from .config_flags import get_config_filename
from .config_flags import get_override_values

__all__ = (
    "DEFINE_config_dataclass",
    "DEFINE_config_dict",
    "DEFINE_config_file",
    "get_config_filename",
    "get_override_values",
)



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Classes for defining configurations of experiments and models."""

from .config_dict import _Op
from .config_dict import ConfigDict
from .config_dict import create
from .config_dict import CustomJSONEncoder
from .config_dict import FieldReference
from .config_dict import FrozenConfigDict
from .config_dict import JSONDecodeError
from .config_dict import MutabilityError
from .config_dict import placeholder
from .config_dict import recursive_rename
from .config_dict import required_placeholder
from .config_dict import RequiredValueError

__all__ = ("_Op", "ConfigDict", "create", "CustomJSONEncoder", "FieldReference",
           "FrozenConfigDict", "JSONDecodeError", "MutabilityError",
           "placeholder", "recursive_rename", "required_placeholder",
           "RequiredValueError")



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Defines a parameterized method which returns a config depending on input."""

import ml_collections


def get_config(config_string):
  """Return an instance of ConfigDict depending on `config_string`."""
  possible_structures = {
      'linear':
          ml_collections.ConfigDict({
              'model_constructor': 'snt.Linear',
              'model_config': ml_collections.ConfigDict({
                  'output_size': 42,
              })
          }),
      'lstm':
          ml_collections.ConfigDict({
              'model_constructor': 'snt.LSTM',
              'model_config': ml_collections.ConfigDict({
                  'hidden_size': 108,
              })
          })
  }

  return possible_structures[config_string]



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Tests for ml_collection.config_flags."""

import copy
import shlex
import sys

from absl import flags
from absl.testing import absltest
from absl.testing import flagsaver
from absl.testing import parameterized
import ml_collections
from ml_collections.config_flags import config_flags
from ml_collections.config_flags.tests import fieldreference_config
from ml_collections.config_flags.tests import mock_config
import six


_CHECK_TYPES = tuple(
    list(six.integer_types) + list(six.string_types) + [float, bool])

_TEST_DIRECTORY = 'ml_collections/config_flags/tests'
_TEST_CONFIG_FILE = '{}/mock_config.py'.format(_TEST_DIRECTORY)
# Parameters to test that config loading and overriding works with both
# one and two dashes.
_DASH_PARAMETERS = (
    ('WithTwoDashesAndEqual', '--test_config={}'.format(_TEST_CONFIG_FILE)),
    ('WithTwoDashes', '--test_config {}'.format(_TEST_CONFIG_FILE)),
    ('WithOneDashAndEqual', '-test_config={}'.format(_TEST_CONFIG_FILE)),
    ('WithOneDash', '-test_config {}'.format(_TEST_CONFIG_FILE)))

_CONFIGDICT_CONFIG_FILE = '{}/configdict_config.py'.format(_TEST_DIRECTORY)
_IOERROR_CONFIG_FILE = '{}/ioerror_config.py'.format(_TEST_DIRECTORY)
_VALUEERROR_CONFIG_FILE = '{}/valueerror_config.py'.format(_TEST_DIRECTORY)
_TYPEERROR_CONFIG_FILE = '{}/typeerror_config.py'.format(_TEST_DIRECTORY)
_FIELDREFERENCE_CONFIG_FILE = '{}/fieldreference_config.py'.format(
    _TEST_DIRECTORY)
_PARAMETERISED_CONFIG_FILE = '{}/parameterised_config.py'.format(
    _TEST_DIRECTORY)


def _parse_flags(command,
                 default=None,
                 config=None,
                 lock_config=True,
                 required=False):
  """Parses arguments simulating sys.argv."""

  if config is not None and default is not None:
    raise ValueError('If config is supplied a default should not be.')

  # Storing copy of the old sys.argv.
  old_argv = list(sys.argv)

  # Overwriting sys.argv, as sys has a global state it gets propagated.
  # The module shlex is useful here because it splits the input similar to
  # sys.argv. For instance, string arguments are not split by space.
  sys.argv = shlex.split(command)

  # Actual parsing.
  values = flags.FlagValues()
  if config is None:
    config_flags.DEFINE_config_file(
        'test_config',
        default=default,
        flag_values=values,
        lock_config=lock_config)
  else:
    config_flags.DEFINE_config_dict(
        'test_config',
        config=config,
        flag_values=values,
        lock_config=lock_config)

  if required:
    flags.mark_flag_as_required('test_config', flag_values=values)
  values(sys.argv)

  # Going back to original values.
  sys.argv = old_argv

  return values


def _get_override_flags(overrides, override_format):
  return ' '.join([override_format.format(path, value)
                   for path, value in six.iteritems(overrides)])


class _ConfigFlagTestCase(object):
  """Base class for tests with additional asserts for comparing configs."""

  def assert_subset_configs(self, config1, config2):
    """Checks if all atrributes/values in config1 are present in config2."""

    if config1 is None:
      return

    if hasattr(config1, '__dict__'):
      keys = [key for key in config1.__dict__ if not key.startswith('_')]
      get_val = getattr
    elif hasattr(config1, 'keys') and callable(config1.keys):
      keys = [key for key in config1.keys() if not key.startswith('_')]
      get_val = lambda haystack, needle: haystack[needle]
    else:
      # This should not fail as it simply means we cannot iterate deeper.
      return

    for attribute in keys:
      if isinstance(get_val(config1, attribute), _CHECK_TYPES):
        self.assertEqual(get_val(config1, attribute),
                         get_val(config2, attribute))
      else:
        # Try to go deeper with comparison
        self.assert_subset_configs(get_val(config1, attribute),
                                   get_val(config2, attribute))

  def assert_equal_configs(self, config1, config2):
    """Checks if two configs are identical."""
    self.assert_subset_configs(config1, config2)
    self.assert_subset_configs(config2, config1)  # pylint: disable=arguments-out-of-order


class ConfigFileFlagTest(_ConfigFlagTestCase, parameterized.TestCase):
  """Tests config flags library."""

  @parameterized.named_parameters(*_DASH_PARAMETERS)
  def testLoading(self, config_flag):
    """Tests loading config from file."""

    values = _parse_flags('./program {}'.format(config_flag),
                          default='nonexisting.py')

    self.assertIn('test_config', values)

    self.assert_equal_configs(values.test_config,
                              mock_config.get_config())

  @parameterized.named_parameters(*_DASH_PARAMETERS)
  def testRequired(self, config_flag):
    """Tests making a config_file flag required."""
    with self.assertRaises(flags.IllegalFlagValueError):
      _parse_flags('./program ', required=True)

    values = _parse_flags('./program {}'.format(config_flag),
                          required=True)
    self.assertIn('test_config', values)

  def testDefaultLoading(self):
    """Tests loading config from file using default path."""

    for required in [True, False]:
      values = _parse_flags(
          './program', default=_TEST_CONFIG_FILE, required=required)

      self.assertIn('test_config', values)

      self.assert_equal_configs(values.test_config,
                                mock_config.get_config())

  def testLoadingNonExistingConfigLoading(self):
    """Tests whether loading non existing file raises an error."""

    nonexisting_file = 'nonexisting.py'

    # Test whether loading non existing files raises an Error with both
    # file loading formats i.e. with '--' and '-'.
    # The Error is not expected to be raised until the config dict actually
    # has one of its attributes accessed.
    values = _parse_flags(
        './program --test_config={}'.format(nonexisting_file),
        default=_TEST_CONFIG_FILE)
    with self.assertRaisesRegex(IOError, '.*{}.*'.format(nonexisting_file)):
      _ = values.test_config.a

    values = _parse_flags(
        './program -test_config {}'.format(nonexisting_file),
        default=_TEST_CONFIG_FILE)
    with self.assertRaisesRegex(IOError, '.*{}.*'.format(nonexisting_file)):
      _ = values.test_config.a

    values = _parse_flags(
        './program -test_config ""', default=_TEST_CONFIG_FILE)
    with self.assertRaisesRegex(IOError, 'empty string'):
      _ = values.test_config.a

  def testIOError(self):
    """Tests that IOErrors raised inside config files are reported correctly."""
    values = _parse_flags('./program --test_config={}'
                          .format(_IOERROR_CONFIG_FILE))
    with self.assertRaisesRegex(IOError, 'This is an IOError'):
      _ = values.test_config.a

  def testValueError(self):
    """Tests that ValueErrors raised when parsing config files are passed up."""
    allchars_regexp = r'[\s\S]*'
    error_regexp = allchars_regexp.join(['Error whilst parsing config file',
                                         'in get_config',
                                         'in value_error_function',
                                         'This is a ValueError'])
    with self.assertRaisesRegex(flags.IllegalFlagValueError, error_regexp):
      _ = _parse_flags('./program --test_config={}'
                       .format(_VALUEERROR_CONFIG_FILE))

  def testTypeError(self):
    """Tests that TypeErrors raised when parsing config files are passed up."""
    allchars_regexp = r'[\s\S]*'
    error_regexp = allchars_regexp.join(['Error whilst parsing config file',
                                         'in get_config',
                                         'in type_error_function',
                                         'This is a TypeError'])
    with self.assertRaisesRegex(flags.IllegalFlagValueError, error_regexp):
      _ = _parse_flags('./program --test_config={}'
                       .format(_TYPEERROR_CONFIG_FILE))

  # Note: While testing the overriding of parameters, we explicitly set
  # '!r' in the format string for the value. This ensures the 'repr()' is
  # called on the argument which basically means that for string arguments,
  # the quotes (' ') are left intact when we format the string.
  @parameterized.named_parameters(
      ('TwoDashConfigAndOverride',
       '--test_config={}'.format(_TEST_CONFIG_FILE), '--test_config.{}={!r}'),
      ('TwoDashSpaceConfigAndOverride',
       '--test_config {}'.format(_TEST_CONFIG_FILE), '--test_config.{} {!r}'),
      ('OneDashConfigAndOverride',
       '-test_config {}'.format(_TEST_CONFIG_FILE), '-test_config.{} {!r}'),
      ('OneDashEqualConfigAndOverride',
       '-test_config={}'.format(_TEST_CONFIG_FILE), '-test_config.{}={!r}'),
      ('OneDashConfigAndTwoDashOverride',
       '-test_config {}'.format(_TEST_CONFIG_FILE), '--test_config.{}={!r}'),
      ('TwoDashConfigAndOneDashOverride',
       '--test_config={}'.format(_TEST_CONFIG_FILE), '-test_config.{} {!r}'))
  def testOverride(self, config_flag, override_format):
    """Tests overriding config values from command line."""
    overrides = {
        'integer': 1,
        'float': -3,
        'dict.float': 3,
        'object.integer': 12,
        'object.float': 123,
        'object.string': 'tom',
        'object.dict.integer': -2,
        'object.dict.float': 3.15,
        'object.dict.list[0]': 101,
        'object.dict.list[2][0]': 103,
        'object.list[0]': 101,
        'object.list[2][0]': 103,
        'object.tuple': '(1,2,(1,2))',
        'object.tuple_with_spaces': '(1, 2, (1, 2))',
        'object_reference.dict.string': 'marry',
        'object.dict.dict.float': 123,
        'object_copy.float': 111.111
    }

    override_flags = _get_override_flags(overrides, override_format)
    values = _parse_flags('./program {} {}'.format(config_flag, override_flags))

    test_config = mock_config.get_config()
    test_config.integer = overrides['integer']
    test_config.float = overrides['float']
    test_config.dict['float'] = overrides['dict.float']
    test_config.object.integer = overrides['object.integer']
    test_config.object.float = overrides['object.float']
    test_config.object.string = overrides['object.string']
    test_config.object.dict['integer'] = overrides['object.dict.integer']
    test_config.object.dict['float'] = overrides['object.dict.float']
    test_config.object.dict['list'][0] = overrides['object.dict.list[0]']
    test_config.object.dict['list'][2][0] = overrides['object.dict.list[2][0]']
    test_config.object.dict['list'][0] = overrides['object.list[0]']
    test_config.object.dict['list'][2][0] = overrides['object.list[2][0]']
    test_config.object.tuple = (1, 2, (1, 2))
    test_config.object.tuple_with_spaces = (1, 2, (1, 2))
    test_config.object_reference.dict['string'] = overrides[
        'object_reference.dict.string']
    test_config.object.dict['dict']['float'] = overrides[
        'object.dict.dict.float']
    test_config.object_copy.float = overrides['object_copy.float']

    self.assert_equal_configs(values.test_config, test_config)

  def testOverrideBoolean(self):
    """Tests overriding boolean config values from command line."""
    prefix = './program --test_config={}'.format(_TEST_CONFIG_FILE)

    # The default for dict.bool is False.
    values = _parse_flags('{} --test_config.dict.bool'.format(prefix))
    self.assertTrue(values.test_config.dict['bool'])

    values = _parse_flags('{} --test_config.dict.bool=true'.format(prefix))
    self.assertTrue(values.test_config.dict['bool'])

    # The default for object.bool is True.
    values = _parse_flags('{} --test_config.object.bool=false'.format(prefix))
    self.assertFalse(values.test_config.object.bool)

    values = _parse_flags('{} --notest_config.object.bool'.format(prefix))
    self.assertFalse(values.test_config.object.bool)

  def testFieldReferenceOverride(self):
    """Tests whether types of FieldReference fields are valid."""
    overrides = {'ref_nodefault': 1, 'ref': 2}
    override_flags = _get_override_flags(overrides, '--test_config.{}={!r}')
    config_flag = '--test_config={}'.format(_FIELDREFERENCE_CONFIG_FILE)
    values = _parse_flags('./program {} {}'.format(config_flag, override_flags))
    cfg = values.test_config

    self.assertEqual(cfg.ref_nodefault, overrides['ref_nodefault'])
    self.assertEqual(cfg.ref, overrides['ref'])

  @parameterized.named_parameters(*_DASH_PARAMETERS)
  def testSetNotExistingKey(self, config_flag):
    """Tests setting value of not existing key."""

    with self.assertRaises(KeyError):
      _parse_flags('./program {} '
                   '--test_config.not_existing_key=1 '.format(config_flag))

  @parameterized.named_parameters(*_DASH_PARAMETERS)
  def testSetReadOnlyField(self, config_flag):
    """Tests setting value of key which is read only."""

    with self.assertRaises(AttributeError):
      _parse_flags('./program {} '
                   '--test_config.readonly_field=1 '.format(config_flag))

  @parameterized.named_parameters(*_DASH_PARAMETERS)
  def testNotSupportedOperation(self, config_flag):
    """Tests setting value to not supported type."""

    with self.assertRaises(config_flags.UnsupportedOperationError):
      _parse_flags('./program {} '
                   '--test_config.list=[1]'.format(config_flag))

  def testReadingNonExistingKey(self):
    """Tests reading non existing key from config."""

    test_config = mock_config.get_config()
    with self.assertRaises(config_flags.UnsupportedOperationError):
      config_flags.SetValue('dict.not_existing_key', test_config, 1)

  def testReadingSettingExistingKeyInDict(self):
    """Tests setting non existing key from dict inside config."""

    test_config = mock_config.get_config()
    with self.assertRaises(KeyError):
      config_flags.SetValue('dict.not_existing_key.key', test_config, 1)

  def testEmptyKey(self):
    """Tests calling an empty key update."""

    test_config = mock_config.get_config()
    with self.assertRaises(ValueError):
      config_flags.SetValue('', test_config, None)

  def testListExtraIndex(self):
    """Tries to index a non-indexable list element."""

    test_config = mock_config.get_config()
    with self.assertRaises(IndexError):
      config_flags.GetValue('dict.list[0][0]', test_config)

  def testListOutOfRangeGet(self):
    """Tries to access out-of-range value in list."""

    test_config = mock_config.get_config()
    with self.assertRaises(IndexError):
      config_flags.GetValue('dict.list[2][1]', test_config)

  def testListOutOfRangeSet(self):
    """Tries to override out-of-range value in list."""

    test_config = mock_config.get_config()
    with self.assertRaises(config_flags.UnsupportedOperationError):
      config_flags.SetValue('dict.list[2][1]', test_config, -1)

  def testParserWrapping(self):
    """Tests callback based Parser wrapping."""

    parser = flags.IntegerParser()

    test_config = mock_config.get_config()
    overrides = {}

    wrapped_parser = config_flags._ConfigFieldParser(parser, 'integer',
                                                     test_config, overrides)

    wrapped_parser.parse('12321')
    self.assertEqual(test_config.integer, 12321)
    self.assertEqual(overrides, {'integer': 12321})

    self.assertEqual(wrapped_parser.flag_type(), parser.flag_type())
    self.assertEqual(wrapped_parser.syntactic_help, parser.syntactic_help)

    self.assertEqual(wrapped_parser.convert('3'), parser.convert('3'))

  def testTypes(self):
    """Tests whether various types of objects are valid."""

    parser = config_flags._ConfigFileParser('test_config')
    self.assertEqual(parser.flag_type(), 'config object')

    test_config = mock_config.get_config()

    paths = (
        'float',
        'integer',
        'string',
        'bool',
        'dict',
        'dict.float',
        'dict.list',
        'list',
        'list[0]',

        'object.float',
        'object.integer',
        'object.string',
        'object.bool',
        'object.dict',
        'object.dict.float',
        'object.dict.list',
        'object.list',
        'object.list[0]',
        'object.tuple',

        'object_reference.float',
        'object_reference.integer',
        'object_reference.string',
        'object_reference.bool',
        'object_reference.dict',
        'object_reference.dict.float',

        'object_copy.float',
        'object_copy.integer',
        'object_copy.string',
        'object_copy.bool',
        'object_copy.dict',
        'object_copy.dict.float'
    )

    paths_types = [
        float,
        int,
        str,
        bool,
        dict,
        float,
        list,
        list,
        int,

        float,
        int,
        str,
        bool,
        dict,
        float,
        list,
        list,
        int,
        tuple,

        float,
        int,
        str,
        bool,
        dict,
        float,

        float,
        int,
        str,
        bool,
        dict,
        float,
    ]

    config_types = config_flags.GetTypes(paths, test_config)

    self.assertEqual(paths_types, config_types)

  def testFieldReferenceTypes(self):
    """Tests whether types of FieldReference fields are valid."""
    test_config = fieldreference_config.get_config()

    paths = ['ref_nodefault', 'ref']
    paths_types = [int, int]

    config_types = config_flags.GetTypes(paths, test_config)
    self.assertEqual(paths_types, config_types)

  @parameterized.named_parameters(
      ('WithTwoDashesAndEqual', '--test_config=config.py'),
      ('WithTwoDashes', '--test_config'),
      ('WithOneDashAndEqual', '-test_config=config.py'),
      ('WithOneDash', '-test_config'))
  def testConfigSpecified(self, config_argument):
    """Tests whether config is specified on the command line."""

    config_flag = config_flags._ConfigFlag(
        parser=flags.ArgumentParser(),
        serializer=None,
        name='test_config',
        default='defaultconfig.py',
        help_string=''
    )
    self.assertTrue(config_flag._IsConfigSpecified([config_argument]))
    self.assertFalse(config_flag._IsConfigSpecified(['']))

  def testFindConfigSpecified(self):
    """Tests whether config is specified on the command line."""

    config_flag = config_flags._ConfigFlag(
        parser=flags.ArgumentParser(),
        serializer=None,
        name='test_config',
        default='defaultconfig.py',
        help_string=''
    )
    self.assertEqual(config_flag._FindConfigSpecified(['']), -1)

    argv_length = 20
    for i in range(argv_length):
      # Generate list of '--test_config.i=0' args.
      argv = ['--test_config.{}=0'.format(arg) for arg in range(argv_length)]
      self.assertEqual(config_flag._FindConfigSpecified(argv), -1)

      # Override i-th arg with something specifying the value of 'test_config'.
      # After doing this, _FindConfigSpecified should return the value of i.
      argv[i] = '--test_config'
      self.assertEqual(config_flag._FindConfigSpecified(argv), i)
      argv[i] = '--test_config=config.py'
      self.assertEqual(config_flag._FindConfigSpecified(argv), i)
      argv[i] = '-test_config'
      self.assertEqual(config_flag._FindConfigSpecified(argv), i)
      argv[i] = '-test_config=config.py'
      self.assertEqual(config_flag._FindConfigSpecified(argv), i)

  def testLoadingLockedConfigDict(self):
    """Tests loading ConfigDict instance and that it is locked."""

    config_flag = '--test_config={}'.format(_CONFIGDICT_CONFIG_FILE)
    values = _parse_flags('./program {}'.format(config_flag),
                          lock_config=True)

    self.assertTrue(values.test_config.is_locked)
    self.assertTrue(values.test_config.nested_configdict.is_locked)

    values = _parse_flags('./program {}'.format(config_flag),
                          lock_config=False)

    self.assertFalse(values.test_config.is_locked)
    self.assertFalse(values.test_config.nested_configdict.is_locked)

  @parameterized.named_parameters(
      ('WithTwoDashesAndEqual', '--test_config={}'.format(_TEST_DIRECTORY)),
      ('WithTwoDashes', '--test_config {}'.format(_TEST_DIRECTORY)),
      ('WithOneDashAndEqual', '-test_config={}'.format(_TEST_DIRECTORY)),
      ('WithOneDash', '-test_config {}'.format(_TEST_DIRECTORY)))
  def testPriorityOfFieldLookup(self, config_flag):
    """Tests whether attributes have higher priority than key-based lookup."""

    values = _parse_flags('./program {}/mini_config.py'.format(config_flag),
                          lock_config=False)
    self.assertTrue(
        config_flags.GetValue('entry_with_collision', values.test_config))

  @parameterized.named_parameters(
      ('TypeAEnabled', ':type_a', {'thing_a': 23, 'thing_b': 42}, ['thing_c']),
      ('TypeASpecify', ':type_a --test_config.thing_a=24', {'thing_a': 24}, []),
      ('TypeBEnabled', ':type_b', {'thing_a': 19, 'thing_c': 65}, ['thing_b']))
  def testExtraConfigString(self, flag_override, should_exist, should_error):
    """Tests with the config_string argument is used properly."""
    values = _parse_flags(
        './program --test_config={}/parameterised_config.py{}'.format(
            _TEST_DIRECTORY, flag_override))
    # Ensure that the values exist in the ConfigDict, with desired values.
    for subfield_name, expected_value in six.iteritems(should_exist):
      self.assertEqual(values.test_config[subfield_name], expected_value)

    # Ensure the values which should not be part of the ConfigDict are really
    # not there.
    for should_error_name in should_error:
      with self.assertRaisesRegex(KeyError, 'Did you mean'):
        _ = values.test_config[should_error_name]

  def testExtraConfigInvalidFlag(self):
    with self.assertRaisesRegex(AttributeError, 'not_valid_item'):
      _parse_flags(
          ('./program --test_config={}/parameterised_config.py:type_a '
           '--test_config.not_valid_item=42').format(_TEST_DIRECTORY))

  def testOverridingConfigDict(self):
    """Tests overriding of ConfigDict fields."""

    config_flag = '--test_config={}'.format(_CONFIGDICT_CONFIG_FILE)
    overrides = {
        'integer': 2,
        'reference': 2,
        'list[0]': 5,
        'nested_list[0][0]': 5,
        'nested_configdict.integer': 5,
        'unusable_config.dummy_attribute': 5
    }
    override_flags = _get_override_flags(overrides, '--test_config.{}={}')
    values = _parse_flags('./program {} {}'.format(config_flag, override_flags))
    self.assertEqual(values.test_config.integer, overrides['integer'])
    self.assertEqual(values.test_config.reference, overrides['reference'])
    self.assertEqual(values.test_config.list[0], overrides['list[0]'])
    self.assertEqual(values.test_config.nested_list[0][0],
                     overrides['nested_list[0][0]'])
    self.assertEqual(values.test_config.nested_configdict.integer,
                     overrides['nested_configdict.integer'])
    self.assertEqual(values.test_config.unusable_config.dummy_attribute,
                     overrides['unusable_config.dummy_attribute'])
    # Attribute error.
    overrides = {'nonexistent': 'value'}
    with self.assertRaises(AttributeError):
      override_flags = _get_override_flags(overrides, '--test_config.{}={}')
      _parse_flags('./program {} {}'.format(config_flag, override_flags))

    # "Did you mean" messages.
    overrides = {'integre': 2}
    with self.assertRaisesRegex(AttributeError, 'Did you.*integer.*'):
      override_flags = _get_override_flags(overrides, '--test_config.{}={}')
      _parse_flags('./program {} {}'.format(config_flag, override_flags))

    overrides = {'referecne': 2}
    with self.assertRaisesRegex(AttributeError, 'Did you.*reference.*'):
      override_flags = _get_override_flags(overrides, '--test_config.{}={}')
      _parse_flags('./program {} {}'.format(config_flag, override_flags))

  # This test adds new flags, so use FlagSaver to make it hermetic.
  @flagsaver.flagsaver
  def testIsConfigFile(self):
    config_flags.DEFINE_config_file('is_a_config_flag')
    flags.DEFINE_integer('not_a_config_flag', -1, '')

    self.assertTrue(
        config_flags.is_config_flag(flags.FLAGS['is_a_config_flag']))
    self.assertFalse(
        config_flags.is_config_flag(flags.FLAGS['not_a_config_flag']))

  # This test adds new flags, so use FlagSaver to make it hermetic.
  @flagsaver.flagsaver
  def testModuleName(self):
    config_flags.DEFINE_config_file('flag')
    argv_0 = './program'
    _parse_flags(argv_0)
    self.assertIn(flags.FLAGS['flag'],
                  flags.FLAGS.flags_by_module_dict()[argv_0])

  def testFlagOrder(self):
    with self.assertRaisesWithLiteralMatch(
        config_flags.FlagOrderError,
        ('Found --test_config.int=1 in argv before a value for --test_config '
         'was specified')):
      _parse_flags('./program --test_config.int=1 '
                   '--test_config={}'.format(_TEST_CONFIG_FILE))

  @flagsaver.flagsaver
  def testOverrideValues(self):
    config_flags.DEFINE_config_file('config')
    with self.assertRaisesWithLiteralMatch(config_flags.UnparsedFlagError,
                                           'The flag has not been parsed yet'):
      flags.FLAGS['config'].override_values  # pylint: disable=pointless-statement

    original_float = -1.0
    original_dictfloat = -2.0
    config = ml_collections.ConfigDict({
        'integer': -1,
        'float': original_float,
        'dict': {
            'float': original_dictfloat
        }
    })
    integer_override = 0
    dictfloat_override = 1.1
    values = _parse_flags('./program --test_config={} --test_config.integer={} '
                          '--test_config.dict.float={}'.format(
                              _TEST_CONFIG_FILE, integer_override,
                              dictfloat_override))

    config.update_from_flattened_dict(
        config_flags.get_override_values(values['test_config']))
    self.assertEqual(config['integer'], integer_override)
    self.assertEqual(config['float'], original_float)
    self.assertEqual(config['dict']['float'], dictfloat_override)

  @parameterized.named_parameters(
      ('ConfigFile1', _TEST_CONFIG_FILE),
      ('ConfigFile2', _CONFIGDICT_CONFIG_FILE),
      ('ParameterisedConfigFile', _PARAMETERISED_CONFIG_FILE + ':type_a'),
      )
  def testConfigPath(self, config_file):
    """Test access to saved config file path."""
    values = _parse_flags('./program --test_config={}'.format(config_file))
    self.assertEqual(config_flags.get_config_filename(values['test_config']),
                     config_file)


def _simple_config():
  config = ml_collections.ConfigDict()
  config.foo = 3
  return config


class ConfigDictFlagTest(_ConfigFlagTestCase, absltest.TestCase):
  """Tests DEFINE_config_dict.

  DEFINE_config_dict reuses a lot of code in DEFINE_config_file so the tests
  here are mostly sanity checks.
  """

  def testBasicUsage(self):
    values = _parse_flags('./program', config=_simple_config())
    self.assertIn('test_config', values)
    self.assert_equal_configs(_simple_config(), values.test_config)

  def testChangingLockedConfigRaisesAnError(self):
    values = _parse_flags(
        './program', config=_simple_config(), lock_config=True)
    with self.assertRaisesRegex(AttributeError, 'config is locked'):
      values.test_config.new_foo = 20

  def testChangingUnlockedConfig(self):
    values = _parse_flags(
        './program', config=_simple_config(), lock_config=False)
    values.test_config.new_foo = 20

  def testNonConfigDictAsConfig(self):
    non_config_dict = dict(a=1, b=2)
    with self.assertRaisesRegex(TypeError, 'should be a ConfigDict'):
      _parse_flags('./program', config=non_config_dict)

  def testOverridingAttribute(self):
    new_foo = 10
    values = _parse_flags(
        './program --test_config.foo={}'.format(new_foo),
        config=_simple_config())
    self.assertNotEqual(new_foo, _simple_config().foo)
    self.assertEqual(new_foo, values.test_config.foo)

  def testOverridingMainConfigFlagRaisesAnError(self):
    with self.assertRaisesRegex(flags.IllegalFlagValueError,
                                'Overriding test_config is not allowed'):
      _parse_flags('./program --test_config=bad_input', config=_simple_config())

  def testOverridesSerialize(self):
    all_types_config = ml_collections.ConfigDict()
    all_types_config.type_bool = False
    all_types_config.type_bytes = b'bytes'
    all_types_config.type_float = 1.0
    all_types_config.type_int = 1
    all_types_config.type_str = 'str'
    all_types_config.type_ustr = u'ustr'
    all_types_config.type_tuple = (False, b'bbytes', 1.0, 1, 'str', u'ustr',)
    # Change values via the command line:
    command_line = (
        './program'
        ' --test_config.type_bool=True'
        ' --test_config.type_float=10'
        ' --test_config.type_int=10'
        ' --test_config.type_str=str_commandline'
        ' --test_config.type_tuple="(\'tuple_str\', 10)"'
        )
    if six.PY3:
      # Not supported in py2; type_bytes is never supported.
      command_line += ' --test_config.type_ustr=ustr_commandline'
    values = _parse_flags(command_line, config=copy.copy(all_types_config))

    # Check we get the expected values (ie the ones defined above not the ones
    # defined in the config itself)
    get_parser = lambda name: values._flags()[name].parser.parse
    get_serializer = lambda name: values._flags()[name].serializer.serialize
    serialize_parse = (
        lambda name, value: get_parser(name)(get_serializer(name)(value)))

    with self.subTest('bool'):
      self.assertNotEqual(values.test_config.type_bool,
                          all_types_config['type_bool'])
      self.assertEqual(values.test_config.type_bool, True)
      self.assertEqual(values.test_config.type_bool,
                       serialize_parse('test_config.type_bool',
                                       values.test_config.type_bool))
    with self.subTest('float'):
      self.assertNotEqual(values.test_config.type_float,
                          all_types_config['type_float'])
      self.assertEqual(values.test_config.type_float, 10.)
      self.assertEqual(values.test_config.type_float,
                       serialize_parse('test_config.type_float',
                                       values.test_config.type_float))
    with self.subTest('int'):
      self.assertNotEqual(values.test_config.type_int,
                          all_types_config['type_int'])
      self.assertEqual(values.test_config.type_int, 10)
      self.assertEqual(values.test_config.type_int,
                       serialize_parse('test_config.type_int',
                                       values.test_config.type_int))
    with self.subTest('str'):
      self.assertNotEqual(values.test_config.type_str,
                          all_types_config['type_str'])
      self.assertEqual(values.test_config.type_str, 'str_commandline')
      self.assertEqual(values.test_config.type_str,
                       serialize_parse('test_config.type_str',
                                       values.test_config.type_str))

    with self.subTest('ustr'):
      if six.PY3:
        self.assertNotEqual(values.test_config.type_ustr,
                            all_types_config['type_ustr'])
        self.assertEqual(values.test_config.type_ustr, u'ustr_commandline')
        self.assertEqual(values.test_config.type_ustr,
                         serialize_parse('test_config.type_ustr',
                                         values.test_config.type_ustr))
    with self.subTest('tuple'):
      self.assertNotEqual(values.test_config.type_tuple,
                          all_types_config['type_tuple'])
      self.assertEqual(values.test_config.type_tuple, ('tuple_str', 10))
      self.assertEqual(values.test_config.type_tuple,
                       serialize_parse('test_config.type_tuple',
                                       values.test_config.type_tuple))


def main():
  absltest.main()


if __name__ == '__main__':
  main()



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
# pylint: disable=line-too-long
r"""Example of basic DEFINE_flag_dict usage.

To run this example with basic config file:
python define_config_dict_basic.py -- \
  --my_config=ml_collections/config_flags/examples/config.py
  \
  --my_config.field1=8 --my_config.nested.field=2.1 \
  --my_config.tuple='(1, 2, (1, 2))'

To run this example with parameterised config file:
python define_config_dict_basic.py -- \
  --my_config=ml_collections/config_flags/examples/parameterised_config.py:linear
  \
  --my_config.model_config.output_size=256'
"""
# pylint: enable=line-too-long

from absl import app

from ml_collections.config_flags import config_flags

_CONFIG = config_flags.DEFINE_config_file('my_config')


def main(_):
  print(_CONFIG.value)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for ml_collections.FrozenConfigDict."""

from collections import abc as collections_abc
import copy
import pickle

from absl.testing import absltest
import ml_collections

_TEST_DICT = {
    'int': 2,
    'list': [1, 2],
    'nested_list': [[1, [2]]],
    'set': {1, 2},
    'tuple': (1, 2),
    'frozenset': frozenset({1, 2}),
    'dict': {
        'float': -1.23,
        'list': [1, 2],
        'dict': {},
        'tuple_containing_list': (1, 2, (3, [4, 5], (6, 7))),
        'list_containing_tuple': [1, 2, [3, 4], (5, 6)],
    },
    'ref': ml_collections.FieldReference({'int': 0})
}


def _test_dict_deepcopy():
  return copy.deepcopy(_TEST_DICT)


def _test_configdict():
  return ml_collections.ConfigDict(_TEST_DICT)


def _test_frozenconfigdict():
  return ml_collections.FrozenConfigDict(_TEST_DICT)


class FrozenConfigDictTest(absltest.TestCase):
  """Tests FrozenConfigDict in config flags library."""

  def assertFrozenRaisesValueError(self, input_list):
    """Assert initialization on all elements of input_list raise ValueError."""
    for initial_dictionary in input_list:
      with self.assertRaises(ValueError):
        _ = ml_collections.FrozenConfigDict(initial_dictionary)

  def testBasicEquality(self):
    """Tests basic equality with different types of initialization."""
    fcd = _test_frozenconfigdict()
    fcd_cd = ml_collections.FrozenConfigDict(_test_configdict())
    fcd_fcd = ml_collections.FrozenConfigDict(fcd)
    self.assertEqual(fcd, fcd_cd)
    self.assertEqual(fcd, fcd_fcd)

  def testImmutability(self):
    """Tests immutability of frozen config."""
    fcd = _test_frozenconfigdict()
    self.assertEqual(fcd.list, tuple(_TEST_DICT['list']))
    self.assertEqual(fcd.tuple, _TEST_DICT['tuple'])
    self.assertEqual(fcd.set, frozenset(_TEST_DICT['set']))
    self.assertEqual(fcd.frozenset, _TEST_DICT['frozenset'])
    # Must manually check set to frozenset conversion, since Python == does not
    self.assertIsInstance(fcd.set, frozenset)

    self.assertEqual(fcd.dict.list, tuple(_TEST_DICT['dict']['list']))
    self.assertNotEqual(fcd.dict.tuple_containing_list,
                        _TEST_DICT['dict']['tuple_containing_list'])
    self.assertEqual(fcd.dict.tuple_containing_list[2][1],
                     tuple(_TEST_DICT['dict']['tuple_containing_list'][2][1]))
    self.assertIsInstance(fcd.dict, ml_collections.FrozenConfigDict)

    with self.assertRaises(AttributeError):
      fcd.newitem = 0
    with self.assertRaises(AttributeError):
      fcd.dict.int = 0
    with self.assertRaises(AttributeError):
      fcd['newitem'] = 0
    with self.assertRaises(AttributeError):
      del fcd.int
    with self.assertRaises(AttributeError):
      del fcd['int']

  def testLockAndFreeze(self):
    """Ensures .lock() and .freeze() raise errors."""
    fcd = _test_frozenconfigdict()

    self.assertFalse(fcd.is_locked)
    self.assertFalse(fcd.as_configdict().is_locked)

    with self.assertRaises(AttributeError):
      fcd.lock()
    with self.assertRaises(AttributeError):
      fcd.unlock()
    with self.assertRaises(AttributeError):
      fcd.freeze()
    with self.assertRaises(AttributeError):
      fcd.unfreeze()

  def testInitConfigDict(self):
    """Tests that ConfigDict initialization handles FrozenConfigDict.

    Initializing a ConfigDict on a dictionary with FrozenConfigDict values
    should unfreeze these values.
    """
    dict_without_fcd_node = _test_dict_deepcopy()
    dict_without_fcd_node.pop('ref')
    dict_with_fcd_node = copy.deepcopy(dict_without_fcd_node)
    dict_with_fcd_node['dict'] = ml_collections.FrozenConfigDict(
        dict_with_fcd_node['dict'])
    cd_without_fcd_node = ml_collections.ConfigDict(dict_without_fcd_node)
    cd_with_fcd_node = ml_collections.ConfigDict(dict_with_fcd_node)
    fcd_without_fcd_node = ml_collections.FrozenConfigDict(
        dict_without_fcd_node)
    fcd_with_fcd_node = ml_collections.FrozenConfigDict(dict_with_fcd_node)

    self.assertEqual(cd_without_fcd_node, cd_with_fcd_node)
    self.assertEqual(fcd_without_fcd_node, fcd_with_fcd_node)

  def testInitCopying(self):
    """Tests that initialization copies when and only when necessary.

    Ensures copying only occurs when converting mutable type to immutable type,
    regardless of whether the FrozenConfigDict is initialized by a dict or a
    FrozenConfigDict. Also ensures no copying occurs when converting from
    FrozenConfigDict back to ConfigDict.
    """
    fcd = _test_frozenconfigdict()

    # These should be uncopied when creating fcd
    fcd_unchanged_from_test_dict = [
        (_TEST_DICT['tuple'], fcd.tuple),
        (_TEST_DICT['frozenset'], fcd.frozenset),
        (_TEST_DICT['dict']['tuple_containing_list'][2][2],
         fcd.dict.tuple_containing_list[2][2]),
        (_TEST_DICT['dict']['list_containing_tuple'][3],
         fcd.dict.list_containing_tuple[3])
    ]

    # These should be copied when creating fcd
    fcd_different_from_test_dict = [
        (_TEST_DICT['list'], fcd.list),
        (_TEST_DICT['dict']['tuple_containing_list'][2][1],
         fcd.dict.tuple_containing_list[2][1])
    ]

    for (x, y) in fcd_unchanged_from_test_dict:
      self.assertEqual(id(x), id(y))
    for (x, y) in fcd_different_from_test_dict:
      self.assertNotEqual(id(x), id(y))

    # Also make sure that converting back to ConfigDict makes no copies
    self.assertEqual(
        id(_TEST_DICT['dict']['tuple_containing_list']),
        id(ml_collections.ConfigDict(fcd).dict.tuple_containing_list))

  def testAsConfigDict(self):
    """Tests that converting FrozenConfigDict to ConfigDict works correctly.

    In particular, ensures that FrozenConfigDict does the inverse of ConfigDict
    regarding type_safe, lock, and attribute mutability.
    """
    # First ensure conversion to ConfigDict works on empty FrozenConfigDict
    self.assertEqual(
        ml_collections.ConfigDict(ml_collections.FrozenConfigDict()),
        ml_collections.ConfigDict())

    cd = _test_configdict()
    cd_fcd_cd = ml_collections.ConfigDict(ml_collections.FrozenConfigDict(cd))
    self.assertEqual(cd, cd_fcd_cd)

    # Make sure locking is respected
    cd.lock()
    self.assertEqual(
        cd, ml_collections.ConfigDict(ml_collections.FrozenConfigDict(cd)))

    # Make sure type_safe is respected
    cd = ml_collections.ConfigDict(_TEST_DICT, type_safe=False)
    self.assertEqual(
        cd, ml_collections.ConfigDict(ml_collections.FrozenConfigDict(cd)))

  def testInitSelfReferencing(self):
    """Ensure initialization fails on self-referencing dicts."""
    self_ref = {}
    self_ref['self'] = self_ref
    parent_ref = {'dict': {}}
    parent_ref['dict']['parent'] = parent_ref
    tuple_parent_ref = {'dict': {}}
    tuple_parent_ref['dict']['tuple'] = (1, 2, tuple_parent_ref)
    attribute_cycle = {'dict': copy.deepcopy(self_ref)}

    self.assertFrozenRaisesValueError(
        [self_ref, parent_ref, tuple_parent_ref, attribute_cycle])

  def testInitCycles(self):
    """Ensure initialization fails if an attribute of input is cyclic."""
    inner_cyclic_list = [1, 2]
    cyclic_list = [3, inner_cyclic_list]
    inner_cyclic_list.append(cyclic_list)
    cyclic_tuple = tuple(cyclic_list)

    test_dict_cyclic_list = _test_dict_deepcopy()
    test_dict_cyclic_tuple = _test_dict_deepcopy()

    test_dict_cyclic_list['cyclic_list'] = cyclic_list
    test_dict_cyclic_tuple['dict']['cyclic_tuple'] = cyclic_tuple

    self.assertFrozenRaisesValueError(
        [test_dict_cyclic_list, test_dict_cyclic_tuple])

  def testInitDictInList(self):
    """Ensure initialization fails on dict and ConfigDict in lists/tuples."""
    list_containing_dict = {'list': [1, 2, 3, {'a': 4, 'b': 5}]}
    tuple_containing_dict = {'tuple': (1, 2, 3, {'a': 4, 'b': 5})}
    list_containing_cd = {'list': [1, 2, 3, _test_configdict()]}
    tuple_containing_cd = {'tuple': (1, 2, 3, _test_configdict())}
    fr_containing_list_containing_dict = {
        'fr': ml_collections.FieldReference([1, {
            'a': 2
        }])
    }

    self.assertFrozenRaisesValueError([
        list_containing_dict, tuple_containing_dict, list_containing_cd,
        tuple_containing_cd, fr_containing_list_containing_dict
    ])

  def testInitFieldReferenceInList(self):
    """Ensure initialization fails on FieldReferences in lists/tuples."""
    list_containing_fr = {'list': [1, 2, 3, ml_collections.FieldReference(4)]}
    tuple_containing_fr = {
        'tuple': (1, 2, 3, ml_collections.FieldReference('a'))
    }

    self.assertFrozenRaisesValueError([list_containing_fr, tuple_containing_fr])

  def testInitInvalidAttributeName(self):
    """Ensure initialization fails on attributes with invalid names."""
    dot_name = {'dot.name': None}
    immutable_name = {'__hash__': None}

    with self.assertRaises(ValueError):
      ml_collections.FrozenConfigDict(dot_name)

    with self.assertRaises(AttributeError):
      ml_collections.FrozenConfigDict(immutable_name)

  def testFieldReferenceResolved(self):
    """Tests that FieldReferences are resolved."""
    cfg = ml_collections.ConfigDict({'fr': ml_collections.FieldReference(1)})
    frozen_cfg = ml_collections.FrozenConfigDict(cfg)
    self.assertNotIsInstance(frozen_cfg._fields['fr'],
                             ml_collections.FieldReference)
    hash(frozen_cfg)  # with FieldReference resolved, frozen_cfg is hashable

  def testFieldReferenceCycle(self):
    """Tests that FieldReferences may not contain reference cycles."""
    frozenset_fr = {'frozenset': frozenset({1, 2})}
    frozenset_fr['fr'] = ml_collections.FieldReference(
        frozenset_fr['frozenset'])
    list_fr = {'list': [1, 2]}
    list_fr['fr'] = ml_collections.FieldReference(list_fr['list'])

    cyclic_fr = {'a': 1}
    cyclic_fr['fr'] = ml_collections.FieldReference(cyclic_fr)
    cyclic_fr_parent = {'dict': {}}
    cyclic_fr_parent['dict']['fr'] = ml_collections.FieldReference(
        cyclic_fr_parent)

    # FieldReference is allowed to point to non-cyclic objects:
    _ = ml_collections.FrozenConfigDict(frozenset_fr)
    _ = ml_collections.FrozenConfigDict(list_fr)
    # But not cycles:
    self.assertFrozenRaisesValueError([cyclic_fr, cyclic_fr_parent])

  def testDeepCopy(self):
    """Ensure deepcopy works and does not affect equality."""
    fcd = _test_frozenconfigdict()
    fcd_deepcopy = copy.deepcopy(fcd)
    self.assertEqual(fcd, fcd_deepcopy)

  def testEquals(self):
    """Tests that __eq__() respects hidden mutability."""
    fcd = _test_frozenconfigdict()

    # First, ensure __eq__() returns False when comparing to other types
    self.assertNotEqual(fcd, (1, 2))
    self.assertNotEqual(fcd, fcd.as_configdict())

    list_to_tuple = _test_dict_deepcopy()
    list_to_tuple['list'] = tuple(list_to_tuple['list'])
    fcd_list_to_tuple = ml_collections.FrozenConfigDict(list_to_tuple)

    set_to_frozenset = _test_dict_deepcopy()
    set_to_frozenset['set'] = frozenset(set_to_frozenset['set'])
    fcd_set_to_frozenset = ml_collections.FrozenConfigDict(set_to_frozenset)

    self.assertNotEqual(fcd, fcd_list_to_tuple)

    # Because set == frozenset in Python:
    self.assertEqual(fcd, fcd_set_to_frozenset)

    # Items are not affected by hidden mutability
    self.assertCountEqual(fcd.items(), fcd_list_to_tuple.items())
    self.assertCountEqual(fcd.items(), fcd_set_to_frozenset.items())

  def testEqualsAsConfigDict(self):
    """Tests that eq_as_configdict respects hidden mutability but not type."""
    fcd = _test_frozenconfigdict()

    # First, ensure eq_as_configdict() returns True with an equal ConfigDict but
    # False for other types.
    self.assertFalse(fcd.eq_as_configdict([1, 2]))
    self.assertTrue(fcd.eq_as_configdict(fcd.as_configdict()))
    empty_fcd = ml_collections.FrozenConfigDict()
    self.assertTrue(empty_fcd.eq_as_configdict(ml_collections.ConfigDict()))

    # Now, ensure it has the same immutability detection as __eq__().
    list_to_tuple = _test_dict_deepcopy()
    list_to_tuple['list'] = tuple(list_to_tuple['list'])
    fcd_list_to_tuple = ml_collections.FrozenConfigDict(list_to_tuple)

    set_to_frozenset = _test_dict_deepcopy()
    set_to_frozenset['set'] = frozenset(set_to_frozenset['set'])
    fcd_set_to_frozenset = ml_collections.FrozenConfigDict(set_to_frozenset)

    self.assertFalse(fcd.eq_as_configdict(fcd_list_to_tuple))
    # Because set == frozenset in Python:
    self.assertTrue(fcd.eq_as_configdict(fcd_set_to_frozenset))

  def testHash(self):
    """Ensures __hash__() respects hidden mutability."""
    list_to_tuple = _test_dict_deepcopy()
    list_to_tuple['list'] = tuple(list_to_tuple['list'])

    self.assertEqual(
        hash(_test_frozenconfigdict()),
        hash(ml_collections.FrozenConfigDict(_test_dict_deepcopy())))
    self.assertNotEqual(
        hash(_test_frozenconfigdict()),
        hash(ml_collections.FrozenConfigDict(list_to_tuple)))

    # Ensure Python realizes FrozenConfigDict is hashable
    self.assertIsInstance(_test_frozenconfigdict(), collections_abc.Hashable)

  def testUnhashableType(self):
    """Ensures __hash__() fails if FrozenConfigDict has unhashable value."""
    unhashable_fcd = ml_collections.FrozenConfigDict(
        {'unhashable': bytearray()})
    with self.assertRaises(TypeError):
      hash(unhashable_fcd)

  def testToDict(self):
    """Ensure to_dict() does not care about hidden mutability."""
    list_to_tuple = _test_dict_deepcopy()
    list_to_tuple['list'] = tuple(list_to_tuple['list'])

    self.assertEqual(_test_frozenconfigdict().to_dict(),
                     ml_collections.FrozenConfigDict(list_to_tuple).to_dict())

  def testPickle(self):
    """Make sure FrozenConfigDict can be dumped and loaded with pickle."""
    fcd = _test_frozenconfigdict()
    locked_fcd = ml_collections.FrozenConfigDict(_test_configdict().lock())

    unpickled_fcd = pickle.loads(pickle.dumps(fcd))
    unpickled_locked_fcd = pickle.loads(pickle.dumps(locked_fcd))

    self.assertEqual(fcd, unpickled_fcd)
    self.assertEqual(locked_fcd, unpickled_locked_fcd)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Setup for pip package."""

from setuptools import find_namespace_packages
from setuptools import setup


def _parse_requirements(requirements_txt_path):
  with open(requirements_txt_path) as fp:
    return fp.read().splitlines()


_VERSION = '0.1.1'


setup(
    name='ml_collections',
    version=_VERSION,
    author='ML Collections Authors',
    author_email='ml-collections@google.com',
    description='ML Collections is a library of Python collections designed for ML usecases.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/google/ml_collections',
    license='Apache 2.0',
    # Contained modules and scripts.
    packages=find_namespace_packages(exclude=['*_test.py']),
    install_requires=_parse_requirements('requirements.txt'),
    tests_require=_parse_requirements('requirements-test.txt'),
    python_requires='>=2.6',
    include_package_data=True,
    zip_safe=False,
    # PyPI package information.
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Configuration commmand line parser."""

import errno
import imp
import os
import re
import sys
import traceback
from typing import Any, Dict, Generic, List, MutableMapping, Optional, Tuple, Type, TypeVar

from absl import flags
from absl import logging
import dataclasses
import ml_collections
from ml_collections.config_flags import tuple_parser
import six
from six import string_types

FLAGS = flags.FLAGS

# We need this to work both for Python 2 and 3.
# The initial code in Python 2 was:
# _FIELD_TYPE_TO_PARSER = {
#     types.IntType: flags.IntegerParser(),
#     types.FloatType: flags.FloatParser(),         # OK For Python 3
#     types.BooleanType: flags.BooleanParser(),     # OK For Python 3
#     types.StringType: flags.ArgumentParser(),
#     types.TupleType: tuple_parser.TupleParser(),  # OK For Python 3
# }
# The possible breaking changes are:
# - A Python 3 int could be a Python 2 long, which was not previously supported.
#   We then add support for long.
# - Only Python 2 str were supported (not unicode). Python 3 will behave the
#   same with the str semantic change.
_FIELD_TYPE_TO_PARSER = {
    float: flags.FloatParser(),
    bool: flags.BooleanParser(),
    # Implementing a custom parser to override `Tuple` arguments.
    tuple: tuple_parser.TupleParser(),
}
for t in six.integer_types:
  _FIELD_TYPE_TO_PARSER[t] = flags.IntegerParser()
for t in six.string_types:
  _FIELD_TYPE_TO_PARSER[t] = flags.ArgumentParser()
_FIELD_TYPE_TO_PARSER[str] = flags.ArgumentParser()
_FIELD_TYPE_TO_SERIALIZER = {
    t: flags.ArgumentSerializer()
    for t in _FIELD_TYPE_TO_PARSER
}


class UnsupportedOperationError(flags.Error):
  pass


class FlagOrderError(flags.Error):
  pass


class UnparsedFlagError(flags.Error):
  pass


def DEFINE_config_file(  # pylint: disable=g-bad-name
    name: str,
    default: Optional[str] = None,
    help_string: str = 'path to config file.',
    flag_values: flags.FlagValues = FLAGS,
    lock_config: bool = True,
    **kwargs) -> flags.FlagHolder:
  r"""Defines flag for `ConfigDict` files compatible with absl flags.

  The flag's value should be a path to a valid python file which contains a
  function called `get_config()` that returns a python object specifying
  a configuration. After the flag is parsed, `FLAGS.name` will contain
  a reference to this object, optionally with some values overridden.

  During flags parsing, every flag of form `--name.([a-zA-Z0-9]+\.?)+=value`
  and `-name.([a-zA-Z0-9]+\.?)+ value` will be treated as an override of a
  specific field in the config object returned by this flag. Field is
  essentially a dot delimited path inside the object where each path element
  has to be either an attribute or a key existing in the config object.
  For example `--my_config.field1.field2=val` means "assign value val
  to the attribute (or key) `field2` inside value of the attribute (or key)
  `field1` inside the value of `my_config` object". If there are both
  attribute and key-based access with the same name, attribute is preferred.

  Typical usage example:

  `script.py`::

    from absl import flags
    from ml_collections.config_flags import config_flags

    FLAGS = flags.FLAGS
    config_flags.DEFINE_config_file('my_config')

    print(FLAGS.my_config)

  `config.py`::

    def get_config():
      return {
          'field1': 1,
          'field2': 'tom',
          'nested': {
              'field': 2.23,
          },
      }

  The following command::

    python script.py -- --my_config=config.py
                        --my_config.field1 8
                        --my_config.nested.field=2.1

  will print::

    {'field1': 8, 'field2': 'tom', 'nested': {'field': 2.1}}

  It is possible to parameterise the get_config function, allowing it to
  return a differently structured result for different occasions. This is
  particularly useful when setting up hyperparameter sweeps across various
  network architectures.

  `parameterised_config.py`::

    def get_config(config_string):
      possible_configs = {
          'mlp': {
              'constructor': 'snt.nets.MLP',
              'config': {
                  'output_sizes': (128, 128, 1),
              }
          },
          'lstm': {
              'constructor': 'snt.LSTM',
              'config': {
                  'hidden_size': 128,
                  'forget_bias': 1.0,
              }
          }
      }
      return possible_configs[config_string]

  If a colon is present in the command line override for the config file,
  everything to the right of the colon is passed into the get_config function.
  The following command lines will both function correctly::

    python script.py -- --my_config=parameterised_config.py:mlp
                        --my_config.config.output_sizes="(256,256,1)"


    python script.py -- --my_config=parameterised_config.py:lstm
                        --my_config.config.hidden_size=256

  The following will produce an error, as the hidden_size flag does not
  exist when the "mlp" config_string is provided::

    python script.py -- --my_config=parameterised_config.py:mlp
                        --my_config.config.hidden_size=256

  Args:
    name: Flag name, optionally including extra config after a colon.
    default: Default value of the flag (default: None).
    help_string: Help string to display when --helpfull is called.
        (default: "path to config file.")
    flag_values: FlagValues instance used for parsing.
        (default: absl.flags.FLAGS)
    lock_config: If set to True, loaded config will be locked through calling
        .lock() method on its instance (if it exists). (default: True)
    **kwargs: Optional keyword arguments passed to Flag constructor.

  Returns:
    a handle to defined flag.
  """
  parser = _ConfigFileParser(name=name, lock_config=lock_config)
  flag = _ConfigFlag(
      parser=parser,
      serializer=flags.ArgumentSerializer(),
      name=name,
      default=default,
      help_string=help_string,
      flag_values=flag_values,
      **kwargs)

  # Get the module name for the frame at depth 1 in the call stack.
  module_name = sys._getframe(1).f_globals.get('__name__', None)  # pylint: disable=protected-access
  module_name = sys.argv[0] if module_name == '__main__' else module_name
  return flags.DEFINE_flag(flag, flag_values, module_name=module_name)


def DEFINE_config_dict(  # pylint: disable=g-bad-name
    name: str,
    config: ml_collections.ConfigDict,
    help_string: str = 'ConfigDict instance.',
    flag_values: flags.FlagValues = FLAGS,
    lock_config: bool = True,
    **kwargs) -> flags.FlagHolder:
  """Defines flag for inline `ConfigDict's` compatible with absl flags.

  Similar to `DEFINE_config_file` except the flag's value should be a
  `ConfigDict` instead of a path to a file containing a `ConfigDict`. After the
  flag is parsed, `FLAGS.name` will contain a reference to the `ConfigDict`,
  optionally with some values overridden.

  Typical usage example:

  `script.py`::

    from absl import flags

    import ml_collections
    from ml_collections.config_flags import config_flags


    config = ml_collections.ConfigDict({
        'field1': 1,
        'field2': 'tom',
        'nested': {
            'field': 2.23,
        }
    })


    FLAGS = flags.FLAGS
    config_flags.DEFINE_config_dict('my_config', config)
    ...

    print(FLAGS.my_config)

  The following command::

    python script.py -- --my_config.field1 8
                        --my_config.nested.field=2.1

  will print::

    field1: 8
    field2: tom
    nested: {field: 2.1}

  Args:
    name: Flag name.
    config: `ConfigDict` object.
    help_string: Help string to display when --helpfull is called.
        (default: "ConfigDict instance.")
    flag_values: FlagValues instance used for parsing.
        (default: absl.flags.FLAGS)
    lock_config: If set to True, loaded config will be locked through calling
        .lock() method on its instance (if it exists). (default: True)
    **kwargs: Optional keyword arguments passed to Flag constructor.

  Returns:
    a handle to defined flag.
  """
  if not isinstance(config, ml_collections.ConfigDict):
    raise TypeError('config should be a ConfigDict')
  parser = _InlineConfigParser(name=name, lock_config=lock_config)
  flag = _ConfigFlag(
      parser=parser,
      serializer=flags.ArgumentSerializer(),
      name=name,
      default=config,
      help_string=help_string,
      flag_values=flag_values,
      **kwargs)

  # Get the module name for the frame at depth 1 in the call stack.
  module_name = sys._getframe(1).f_globals.get('__name__', None)  # pylint: disable=protected-access
  module_name = sys.argv[0] if module_name == '__main__' else module_name
  return flags.DEFINE_flag(flag, flag_values, module_name=module_name)


# Note that we would add a bound to constrain this to be a dataclass, except
# that dataclasses don't have a specific base class, and structural typing for
# attributes is currently (2021Q1) not supported in pytype (b/150927776).
_T = TypeVar('_T')


class _TypedFlagHolder(flags.FlagHolder, Generic[_T]):
  """A typed wrapper for a FlagHolder."""

  def __init__(self, flag: flags.FlagHolder, ctor: Type[_T]):
    self._flag = flag
    self._ctor = ctor

  @property
  def value(self) -> _T:
    return self._ctor(**self._flag.value)

  @property
  def default(self) -> _T:
    return self._ctor(**self._flag.default)

  @property
  def name(self) -> str:
    return self._flag.name


def DEFINE_config_dataclass(  # pylint: disable=invalid-name
    name: str,
    config: _T,
    help_string: str = 'Configuration object. Must be a dataclass.',
    flag_values: flags.FlagValues = FLAGS,
    **kwargs,
) -> _TypedFlagHolder[_T]:
  """Defines a typed (dataclass) flag-overrideable configuration.

  Similar to `DEFINE_config_dict` except `config` should be a `dataclass`.

  Args:
    name: Flag name.
    config: A user-defined configuration object. Must be built via `dataclass`.
    help_string: Help string to display when --helpfull is called.
    flag_values: FlagValues instance used for parsing.
    **kwargs: Optional keyword arguments passed to Flag constructor.
  Returns:
    A handle to the defined flag.
  """

  if not dataclasses.is_dataclass(config):
    raise ValueError('Configuration object must be a `dataclass`.')

  # Convert to configdict *without* recursing into leaf node(s).
  # If our config contains dataclasses (or other types) as fields, we want to
  # preserve them; dataclasses.asdict recursively turns all fields into dicts.
  dictionary = {field.name: getattr(config, field.name)
                for field in dataclasses.fields(config)}
  config_dict = ml_collections.ConfigDict(initial_dictionary=dictionary)

  # Define the flag.
  config_flag = DEFINE_config_dict(
      name,
      config=config_dict,
      help_string=help_string,
      flag_values=flag_values,
      **kwargs,
  )

  return _TypedFlagHolder(flag=config_flag, ctor=config.__class__)


def get_config_filename(config_flag) -> str:  # pylint: disable=g-bad-name
  """Returns the path to the config file given the config flag.

  Args:
    config_flag: The flag instance obtained from FLAGS, e.g. FLAGS['config'].

  Returns:
    the path to the config file.
  """
  if not is_config_flag(config_flag):
    raise TypeError('expect a config flag, found {}'.format(type(config_flag)))
  return config_flag.config_filename


def get_override_values(config_flag) -> Dict[str, Any]:  # pylint: disable=g-bad-name
  """Returns a flat dict containing overridden values from the config flag.

  Args:
    config_flag: The flag instance obtained from FLAGS, e.g. FLAGS['config'].

  Returns:
    a flat dict containing overridden values from the config flag.
  """
  if not is_config_flag(config_flag):
    raise TypeError('expect a config flag, found {}'.format(type(config_flag)))
  return config_flag.override_values


class _IgnoreFileNotFoundAndCollectErrors:
  """Helps recording "file not found" exceptions when loading config.

  Usage:
    ignore_errors = _IgnoreFileNotFoundAndCollectErrors()
    with ignore_errors.Attempt('Loading from foo', 'bar.id'):
      ...
      return True  # successfully loaded from `foo`
    logging.error('Failed loading: {}'.format(ignore_errors.DescribeAttempts()))
  """

  def __init__(self):
    self._attempts = []  # type: List[Tuple[Tuple[str, str], IOError]]

  def Attempt(self, description, path):
    """Creates a context manager that routes exceptions to this class."""
    self._current_attempt = (description, path)
    ignore_errors = self

    class _ContextManager:

      def __enter__(self):
        return self

      def __exit__(self, exc_type, exc_value, unused_traceback):
        return ignore_errors.ProcessAttemptException(exc_type, exc_value)

    return _ContextManager()

  def ProcessAttemptException(self, exc_type, exc_value):
    expected_type = IOError if six.PY2 else FileNotFoundError  # pylint: disable=undefined-variable
    if exc_type is expected_type and exc_value.errno == errno.ENOENT:
      self._attempts.append((self._current_attempt, exc_value))
      # Returning a true value suppresses exceptions:
      # https://docs.python.org/2/reference/datamodel.html#object.__exit__
      return True

  def DescribeAttempts(self):
    return '\n'.join(
        '  Attempted [{}]:\n    {}\n      {}'.format(attempt[0], attempt[1], e)
        for attempt, e in self._attempts)


def _LoadConfigModule(name: str, path: str):
  """Loads a script from external file specified by path.

  Unprefixed path is looked for in the current working directory using
  regular file open operation. This should work with relative config paths.

  Args:
    name: Name of the new module.
    path: Path to the .py file containing the module.

  Returns:
    Module loaded from the given path.

  Raises:
    IOError: If the config file cannot be found.
  """
  if not path:
    raise IOError('Path to config file is an empty string.')

  ignoring_errors = _IgnoreFileNotFoundAndCollectErrors()

  # Works for relative paths.
  with ignoring_errors.Attempt('Relative path', path):
    config_module = imp.load_source(name, path)
    return config_module

  # Nothing worked. Log the paths that were attempted.
  raise IOError('Failed loading config file {}\n{}'.format(
      name, ignoring_errors.DescribeAttempts()))


class _ErrorConfig:
  """Dummy ConfigDict that raises an error on any attribute access."""

  def __init__(self, error):
    super(_ErrorConfig, self).__init__()
    super(_ErrorConfig, self).__setattr__('_error', error)

  def __getattr__(self, attr):
    self._ReportError()

  def __setattr__(self, attr, value):
    self._ReportError()

  def __delattr__(self, attr):
    self._ReportError()

  def __getitem__(self, key):
    self._ReportError()

  def __setitem__(self, key, value):
    self._ReportError()

  def __delitem__(self, key):
    self._ReportError()

  def _ReportError(self):
    raise IOError('Configuration is not available because of an earlier '
                  'failure to load: ' +
                  # 'message' is not available in Python 3.
                  getattr(self._error, 'message', str(self._error)))


def _LockConfig(config):
  """Calls config.lock() if config has a lock method."""
  if isinstance(config, _ErrorConfig):
    pass  # Attempting to access _ErrorConfig.lock will raise its error.
  elif getattr(config, 'lock', None) and callable(config.lock):
    config.lock()
  else:
    pass  # config.lock() does not have desired semantics, do nothing.


class _ConfigFileParser(flags.ArgumentParser):
  """Parser for config files."""

  def __init__(self, name, lock_config=True):
    self.name = name
    self._lock_config = lock_config

  def parse(self, path):
    """Loads a config module from `path` and returns the `get_config()` result.

    If a colon is present in `path`, everything to the right of the first colon
    is passed to `get_config` as an argument. This allows the structure of what
    is returned to be modified, which is useful when performing complex
    hyperparameter sweeps.

    Args:
      path: string, path pointing to the config file to execute. May also
          contain a config_string argument, e.g. be of the form
          "config.py:some_configuration".
    Returns:
      Result of calling `get_config` in the specified module.
    """
    # This will be a 2 element list iff extra configuration args are present.
    split_path = path.split(':', 1)
    try:
      config_module = _LoadConfigModule('{}_config'.format(self.name),
                                        split_path[0])
      config = config_module.get_config(*split_path[1:])
      if config is None:
        logging.warning(
            '%s:get_config() returned None, did you forget a return statement?',
            path)
    except IOError as e:
      # Don't raise the error unless/until the config is actually accessed.
      config = _ErrorConfig(e)
    # Third party flags library catches TypeError and ValueError and rethrows,
    # removing useful information unless it is added here (b/63877430):
    except (TypeError, ValueError) as e:
      error_trace = traceback.format_exc()
      raise type(e)('Error whilst parsing config file:\n\n' + error_trace)

    if self._lock_config:
      _LockConfig(config)

    return config

  def flag_type(self):
    return 'config object'


class _InlineConfigParser(flags.ArgumentParser):
  """Parser for a config defined inline (not from a file)."""

  def __init__(self, name, lock_config=True):
    self.name = name
    self._lock_config = lock_config

  def parse(self, config):
    if not isinstance(config, ml_collections.ConfigDict):
      raise TypeError('Overriding {} is not allowed.'.format(self.name))
    if self._lock_config:
      _LockConfig(config)
    return config

  def flag_type(self):
    return 'config object'


class _ConfigFlag(flags.Flag):
  """Flag definition for command-line overridable configs."""

  def __init__(self, flag_values=FLAGS, **kwargs):
    # Parent constructor can already call .Parse, thus additional fields
    # have to be set here.
    self.flag_values = flag_values
    super(_ConfigFlag, self).__init__(**kwargs)

  def _GetOverrides(self, argv):
    """Parses the command line arguments for the overrides."""
    overrides = []
    config_index = self._FindConfigSpecified(argv)
    for i, arg in enumerate(argv):
      if re.match(r'-{{1,2}}(no)?{}\.'.format(self.name), arg):
        if config_index > 0 and i < config_index:
          raise FlagOrderError('Found {} in argv before a value for --{} '
                               'was specified'.format(arg, self.name))
        arg_name = arg.split('=', 1)[0]
        overrides.append(arg_name.split('.', 1)[1])
    return overrides

  def _FindConfigSpecified(self, argv):
    """Finds element in argv specifying the value of the config flag.

    Args:
      argv: command line arguments as a list of strings.
    Returns:
      Index in argv if found and -1 otherwise.
    """
    for i, arg in enumerate(argv):
      # '-(-)config' followed by '=' or at the end of the string.
      if re.match(r'^-{{1,2}}{}(=|$)'.format(self.name), arg) is not None:
        return i
    return -1

  def _IsConfigSpecified(self, argv):
    """Returns `True` if the config file is specified on the command line."""
    return self._FindConfigSpecified(argv) >= 0

  def _set_default(self, default):
    if self._IsConfigSpecified(sys.argv):
      self.default = default
    else:
      super(_ConfigFlag, self)._set_default(default)  # pytype: disable=attribute-error
    self.default_as_str = "'{}'".format(default)

  def _parse(self, argument):
    # Parse config
    config = super(_ConfigFlag, self)._parse(argument)

    # Get list or overrides
    overrides = self._GetOverrides(sys.argv)
    # Attach types definitions
    overrides_types = GetTypes(overrides, config)

    # Iterate over overridden fields and create valid parsers
    self._override_values = {}
    for field_path, field_type in zip(overrides, overrides_types):
      field_help = 'An override of {}\'s field {}'.format(self.name, field_path)
      field_name = '{}.{}'.format(self.name, field_path)

      if field_type in _FIELD_TYPE_TO_PARSER:
        parser = _ConfigFieldParser(_FIELD_TYPE_TO_PARSER[field_type],
                                    field_path, config, self._override_values)
        flags.DEFINE(
            parser,
            field_name,
            GetValue(field_path, config),
            field_help,
            flag_values=self.flag_values,
            serializer=_FIELD_TYPE_TO_SERIALIZER[field_type])
        flag = self.flag_values._flags().get(field_name)  # pylint: disable=protected-access
        flag.boolean = field_type is bool
      else:
        raise UnsupportedOperationError(
            "Type {} of field {} is not supported for overriding. "
            "Currently supported types are: {}. (Note that tuples should "
            "be passed as a string on the command line: flag='(a, b, c)', "
            "rather than flag=(a, b, c).)".format(
                field_type, field_name, _FIELD_TYPE_TO_PARSER.keys()))

    self._config_filename = argument
    return config

  @property
  def config_filename(self):
    """Returns a path to a config file.

    Typical usage example:
    `script.py`:

    ```python
    ...
    from absl import flags
    from ml_collections.config_flags import config_flags

    FLAGS = flags.FLAGS
    config_flags.DEFINE_config_file(
      name='my_config',
      default='ml_collections/config_flags/tests/configdict_config.py',
      help_string='config file')
    ...

    FLAGS['my_config'].config_filename

    will output
    'ml_collections/config_flags/tests/configdict_config.py'
    ```

    Returns:
      A path to a config file. For a parameterised get_config, the config
      filename with the provided parameterisation is returned.

    Raises:
      UnparsedFlagError: if the flag has not been parsed.

    """
    if not hasattr(self, '_config_filename'):
      raise UnparsedFlagError('The flag has not been parsed yet')
    return self._config_filename

  @property
  def override_values(self):
    """Returns a flat dictionary containing overridden values.

    Keys in the dictionary are dot-separated paths navigating to child items in
    the original configuration. For example, supppose that a `config` flag is
    defined and initialized to the following configuration:

    ```python
    {
        'a': 1,
        'nested': {
            'b': 2
        }
    }
    ```

    and the user overrides both values using command-line flags:

    ```
    --config.a=10 --config.nested.b=20
    ```

    Then `FLAGS['config'].override_values` will return:

    ```python
    {
        'a': 10,
        'nested.b': 20
    }
    ```

    The result can be passed to `ConfigDict.update_from_flattened_dict` to
    update the values in a configuration. Continuing with the example above:

    ```python
    import ml_collections
    config = ml_collections.ConfigDict{
        'a': 123,
        'nested': {
            'b': 456
        }
    }
    config.update_from_flattened_dict(FLAGS['config'].override_values)
    print(config.a)  # Prints `10`.
    print(config.nested.b)  # Prints `20`.
    ```

    Returns:
      Flat dictionary with overridden values.

    Raises:
      UnparsedFlagError: if the flag has not been parsed.
    """
    if not hasattr(self, '_override_values'):
      raise UnparsedFlagError('The flag has not been parsed yet')
    return self._override_values


def is_config_flag(flag):  # pylint: disable=g-bad-name
  """Returns True iff `flag` is an instance of `_ConfigFlag`.

  External users of the library may need to check if a flag is of this type
  or not, particularly because ConfigFlags should be parsed before any other
  flags. This function allows that test to be done without making the whole
  class public.

  Args:
    flag: Flag object.

  Returns:
    True iff `isinstance(flag, _ConfigFlag)` is true.
  """
  return isinstance(flag, _ConfigFlag)


class _ConfigFieldParser(flags.ArgumentParser):
  """Parser with config update after parsing.

  This class-based wrapper creates a new object, which uses
  existing parser to do actual parsing and attaches a single callback to
  SetValue afterwards, which is used to update a predefined path in
  the config object.
  """

  def __init__(
      self,
      parser: flags.ArgumentParser,
      path: str,
      config: ml_collections.ConfigDict,
      override_values: MutableMapping[str, Any]):
    """Creates new parser with callback, using existing one to perform parsing.

    Args:
      parser: ArgumentParser instance to wrap.
      path: Dot separated path in config to update with the result of
          parser.parse(...)
      config: Reference to the config object.
      override_values: Dictionary with override values. The 'parse' method will
          add the parsed value to this dictionary with key `path`.
    """
    self._parser = parser
    self._path = path
    self._config = config
    self._override_values = override_values

  def __getattr__(self, attr):
    return getattr(self._parser, attr)

  def parse(self, argument):  # pylint: disable=invalid-name
    value = self._parser.parse(argument)
    SetValue(self._path, self._config, value)
    self._override_values[self._path] = value
    return value

  def flag_type(self) -> str:
    return self._parser.flag_type()

  @property
  def syntactic_help(self) -> str:
    return self._parser.syntactic_help


def _ExtractIndicesFromStep(step):
  """Separates identifier from indexes. Example: id[1][10] -> 'id', [1, 10]."""
  # Map returns an iterable in Python 3, thus the additional `list`.
  return step.split('[', 1)[0], list(map(int, re.findall(r'\[(\d+)\]', step)))


def _AccessConfig(current, field, indices):
  """Returns member of the current config.

  The field in `current` specified by the `field` parameter is indexed using the
  values in the `indices` indices.

  Example:

  ```python
  current = {'first': [0], 'second': [[1]]}
  _AccessConfig(current, 'first', [0]) # Returns 0
  _AccessConfig(current, 'second', [0, 0]) # Returns 1

  ```

  Args:
    current: Current config.
    field: Field to access (string).
    indices: List of indices.

  Returns:
    Member of the config.

  Raises:
    IndexError: if indices are invalid.
    KeyError: if the field is not found.
  """
  if isinstance(field, string_types) and hasattr(current, field):
    current = getattr(current, field)
  elif hasattr(current, '__getitem__'):
    current = current[field]
  else:
    raise KeyError(field)

  for i, index in enumerate(indices):
    try:
      current = current[index]
    except TypeError:
      msg = 'Could not index config {}'.format(field)
      if i > 0:
        msg += '[{}]'.format(']['.join(map(str, indices[:i])))
      raise IndexError(msg)
    except IndexError:
      msg = 'Index [{}] not found in config {}'.format(
          ']['.join(map(str, indices[i:])), field)
      if i > 0:
        msg += '[{}]'.format(']['.join(map(str, indices[:i])))
      raise IndexError(msg)

  return current


def _TakeStep(current, step):
  field, indices = _ExtractIndicesFromStep(step)
  return _AccessConfig(current, field, indices)


def GetValue(path: str, config: ml_collections.ConfigDict):
  """Gets value of a single field."""
  current = config
  for step in path.split('.'):
    current = _TakeStep(current, step)
  return current


def GetType(path, config):
  """Gets type of field in config described by a dotted delimited path."""
  steps = path.split('.')
  current = config
  for step in steps[:-1]:
    current = _TakeStep(current, step)

  # We need to check if there is list-type indexing in the last step.
  field, indices = _ExtractIndicesFromStep(steps[-1])
  if indices:
    return type(_AccessConfig(current, field, indices))
  else:
    # Check if current is a DM collection and hence has attribute get_type()
    is_dm_collection = isinstance(current,
                                  ml_collections.ConfigDict) or isinstance(
                                      current, ml_collections.FieldReference)
    if is_dm_collection:
      return current.get_type(steps[-1])
    else:
      return type(_TakeStep(current, steps[-1]))


def GetTypes(paths, config):
  """Gets types of fields in config described by dotted delimited paths."""
  return [GetType(path, config) for path in paths]


def SetValue(path: str, config: ml_collections.ConfigDict, value):
  """Sets value of a single field."""
  current = config
  steps = path.split('.')

  if not steps or not path:
    raise ValueError('Path cannot be empty')

  for step in steps[:-1]:
    current = _TakeStep(current, step)

  field, indices = _ExtractIndicesFromStep(steps[-1])
  if indices:
    current = _AccessConfig(current, field, indices[:-1])
    try:
      current[indices[-1]] = value
    except:
      raise UnsupportedOperationError(
          'Config does not have a setter for {}'.format(path))
  else:
    if hasattr(current, '__setitem__') and field in current:
      current[field] = value
    elif hasattr(current, field):
      setattr(current, field, value)
    else:
      raise UnsupportedOperationError(
          'Config does not have a setter for {}'.format(path))



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for config_flags examples.

Ensures that from define_config_dict_basic, define_config_file_basic run
successfully.
"""

from absl import flags
from absl.testing import absltest
from absl.testing import flagsaver
from ml_collections.config_flags.examples import define_config_dict_basic
from ml_collections.config_flags.examples import define_config_file_basic

FLAGS = flags.FLAGS


class ConfigDictExamplesTest(absltest.TestCase):

  def test_define_config_dict_basic(self):
    define_config_dict_basic.main([])

  @flagsaver.flagsaver
  def test_define_config_file_basic(self):
    FLAGS.my_config = 'ml_collections/config_flags/examples/config.py'
    define_config_file_basic.main([])

  @flagsaver.flagsaver
  def test_define_config_file_parameterised(self):
    FLAGS.my_config = 'ml_collections/config_flags/examples/parameterised_config.py:linear'
    define_config_file_basic.main([])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for config_flags used in conjunction with DEFINE_config_dataclass."""

import shlex
import sys
from typing import Mapping, Optional, Sequence

from absl import flags
from absl.testing import absltest
import dataclasses
from ml_collections import config_flags


#####
# Simple dummy configuration classes.
@dataclasses.dataclass
class MyModelConfig:
  foo: int
  bar: Sequence[str]
  baz: Optional[Mapping[str, str]] = None


@dataclasses.dataclass
class MyConfig:
  my_model: MyModelConfig
  baseline_model: MyModelConfig


_CONFIG = MyConfig(
    my_model=MyModelConfig(
        foo=3,
        bar=['a', 'b'],
        baz={'foo': 'bar'},
    ),
    baseline_model=MyModelConfig(
        foo=55,
        bar=['c', 'd'],
    ),
)

# Define the flag.
_CONFIG_FLAG = config_flags.DEFINE_config_dataclass('config', _CONFIG)


class TypedConfigFlagsTest(absltest.TestCase):

  def test_instance(self):
    config = _CONFIG_FLAG.value
    self.assertIsInstance(config, MyConfig)
    self.assertEqual(config.my_model, _CONFIG.my_model)
    self.assertEqual(_CONFIG, config)

  def test_flag_overrides(self):

    # Set up some flag overrides.
    old_argv = list(sys.argv)
    sys.argv = shlex.split(
        './program foo.py --test_config.baseline_model.foo=99')
    flag_values = flags.FlagValues()

    # Define a config dataclass flag.
    test_config = config_flags.DEFINE_config_dataclass(
        'test_config', _CONFIG, flag_values=flag_values)

    # Inject the flag overrides.
    flag_values(sys.argv)
    sys.argv = old_argv

    # Did the value get overridden?
    self.assertEqual(test_config.value.baseline_model.foo, 99)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The ML Collections Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Example of ConfigDict usage of lock.

This example shows the roles and scopes of ConfigDict's lock().
"""

from absl import app
import ml_collections


def main(_):
  cfg = ml_collections.ConfigDict()
  cfg.integer_field = 123

  # Locking prohibits the addition and deletion of new fields but allows
  # modification of existing values. Locking happens automatically during
  # loading through flags.
  cfg.lock()
  try:
    cfg.intagar_field = 124  # Raises AttributeError and suggests valid field.
  except AttributeError as e:
    print(e)
  cfg.integer_field = -123  # Works fine.

  with cfg.unlocked():
    cfg.intagar_field = 1555  # Works fine too.

  print(cfg)


if __name__ == '__main__':
  app.run()
